
# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2021-10-10T02:14:13+0000",
 "dirty": false,
 "error": null,
 "full-revisionid": "b9730b6db93f1c5c18d3237f1ac6b8afa9a067f0",
 "version": "0.8.dev2141"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
